/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Jasper van de Gronde (version and extra printf)
 */

/* Jasper van de Gronde
 * Added support for running as a service on Win32.
 */
/**
 * \todo Languages and datatypes? (not needed?)
 * \todo Rename URL to URI and perhaps similar cosmetic changes. (anonymous resources->blank nodes)
 * \todo See http://www.w3.org/TR/2004/REC-rdf-mt-20040210/ (section 3.3.2), containers
 *       should be handled according to those specifications.
 */

#include "rdf-int.h"
#include <stdio.h>
#include "../utils/utils.h"
#ifdef WIN32
#include <windows.h>
#endif

/* Initializes utils.c */
int initUtils(int argc, char* argv[]);
/* Initializes all parts of RDFDB (except for utils.c). Returns the number of init's that failed. */
int initAll();
/* Opens the server part, doesn't return very soon. Returns a non-zero value it if fails, a zero value if it succeeds. */
int openService();
/* Shuts all parts down. Returns the number of deiniters that failed. */
int deinitAll();
/* Deinitializes utils.c */
int deinitUtils();

#define VERSION "0.47 (Sept 08 2001)"
/*
#define VERSION "0.47 (Sept 08 2001), modified April 15 2004"
#define VERSION "0.47 (Sept 08 2001), modified March 19 2004 (2nd)"
#define VERSION "0.47 (Sept 08 2001), modified March 18 2004"
#define VERSION "0.47 (Sept 08 2001), modified March 14 2004"
#define VERSION "0.47 (Sept 08 2001), modified March 10 2004"
#define VERSION "0.47 (Sept 08 2001), modified Feb 13 2004"
*/

#ifdef WIN32
SERVICE_STATUS          RDFDBServiceStatus; 
SERVICE_STATUS_HANDLE   RDFDBServiceStatusHandle; 

void WINAPI RDFDBServiceStart(DWORD argc, LPTSTR *argv);
void WINAPI RDFDBServiceCtrlHandler(DWORD opcode); 
void CreateRDFDBService(int create, int argc, char* argv[]);
#endif

int
main (int argc, char* argv[]) {
  PrintMsg(NOTICE, "\nrdfDB server %s, modified " __DATE__ ".\n", VERSION);
  PrintMsg(NOTICE, "For more information see http://www.guha.com/rdfdb/\n");
  PrintMsg(NOTICE, "And see http://home.hccnet.nl/th.v.d.gronde/dev/rdfdb/ for information about the modifications and the win32 port.\n");

  initUtils(argc, argv); /* Called separately because some of these things are already needed before initAll is called. */

  {
    time_t tm ;
    tm = time(&tm);
    PrintMsg(INFO, "\n%sRDFDB Starting.\n\n", ctime(&tm));
  }

  {
#ifdef WIN32
    const char* service_setting = get_setting("SERVICE");
    if ( service_setting ) {
      if ( strcasecmp(service_setting, "RUN") == 0 ) {
        SERVICE_TABLE_ENTRY DispatchTable[] = 
        { 
          {TEXT("RDFDB"), RDFDBServiceStart}, 
          {0,0} 
        }; 

        if (!StartServiceCtrlDispatcher(DispatchTable)) 
        { 
          PrintMsg(CRITICAL, "StartServiceCtrlDispatcher error = %d\n", GetLastError()); 
        }
      }
      else if ( strcasecmp(service_setting, "INSTALL") == 0 ) {
        CreateRDFDBService(1, argc, argv);
      }
      else if ( strcasecmp(service_setting, "UNINSTALL") == 0 ) {
        CreateRDFDBService(0, 0, 0);
      }
      else {
        PrintMsg(CRITICAL, "Invalid service option: %s\n", service_setting);
      }
    }
    else
#endif
    {
      int failures;
      if ( 0 != (failures=initAll()) ) {
        PrintMsg(CRITICAL, "%i initializers failed.\n", failures);
      }
      else {
        openService(); 
      }
      if ( 0 != (failures=deinitAll()) ) {
        PrintMsg(CRITICAL, "%i deinitializers failed.\n", failures);
      }
    }
  }

  {
    time_t tm ;
    tm = time(&tm);
    PrintMsg(INFO, "\n%sRDFDB Exited.\n", ctime(&tm));
  }

  deinitUtils();

  return 0;
}

/* This is what the (de)initialize structure looks like:
      RDF_Initialize
        initIndexPool
        DB_Initialize
      RDF_Deinitialize
        deinitIndexPool
        DB_Deinitialize
          EmptyCursorPool
          CloseAllDatabases

      RDF_Initparsers
      RDF_Deinitparsers

      initProcessQuery
      deinitProcessQuery

      initQueryParser
      deinitQueryParser

      initServer
        init_connection
      deinitServer
        deinit_connection

      init_security
      deinit_security

      initUtils
      deinitUtils
   Top level functions are called by the initAll and deinitAll functions below (except for the (de)initUtils functions, those are called from main).
*/

int initParsers();
int initServer();
int init_security();
int initQueryParser();
int initProcessQuery();

/* JvdG */
/** Does NOT call initUtils. */
int initAll(int argc, char* argv[]) {
  int failures = 0;
  /* Called elsewhere: if ( 0 != initUtils(argc, argv) ) ++failures; */
  if ( 0 != RDF_Initialize() ) ++failures; /* Also takes care of DB_Initialize and initIndexPool */
  if ( 0 != initParsers() ) ++failures;
  if ( 0 != initQueryParser() ) ++failures;
  if ( 0 != initProcessQuery() ) ++failures;
  if ( 0 != init_security() ) ++failures;
  if ( 0 != initServer() ) ++failures;
  return failures;
}

int deinitParsers();
int deinitServer();
int deinit_security();
int deinitQueryParser();
int deinitProcessQuery();

/* JvdG */
/** Does NOT call deinitUtils or deinit. */
int deinitAll() {
  int failures = 0;
  if ( 0 != deinitServer() ) ++failures;
  if ( 0 != deinit_security() ) ++failures;
  if ( 0 != RDF_Deinitialize() ) ++failures; /* Also takes care of DB_Deinitialize and deinitIndexPool */
  if ( 0 != deinitParsers() ) ++failures;
  if ( 0 != deinitQueryParser() ) ++failures;
  if ( 0 != deinitProcessQuery() ) ++failures;
  /* Called elsewhere: if ( 0 != deinitUtils() ) ++failures; */
  return failures;
}

#ifdef WIN32
/* JvdG */
void WINAPI RDFDBServiceStart(DWORD argc, LPTSTR *argv) 
{ 
  RDFDBServiceStatus.dwServiceType        = SERVICE_WIN32; /* SERVICE_WIN32_OWN_PROCESS? */
  RDFDBServiceStatus.dwCurrentState       = SERVICE_START_PENDING; 
  RDFDBServiceStatus.dwControlsAccepted   = SERVICE_ACCEPT_STOP; 
  RDFDBServiceStatus.dwWin32ExitCode      = NO_ERROR; 
  RDFDBServiceStatus.dwServiceSpecificExitCode = 0; 
  RDFDBServiceStatus.dwCheckPoint         = 0; 
  RDFDBServiceStatus.dwWaitHint           = 500; /* The maximum number of milliseconds between increments of dwCheckPoint. */

  RDFDBServiceStatusHandle = RegisterServiceCtrlHandler(
    TEXT("RDFDB"),
    RDFDBServiceCtrlHandler);

  if ( RDFDBServiceStatusHandle == 0 ) { 
    PrintMsg(CRITICAL, "RegisterServiceCtrlHandler failed %d\n", GetLastError()); 
    return; 
  } 

  {
    int failures;
    if ( 0 != (failures=initAll()) ) {
      PrintMsg(CRITICAL, "%i initializers failed.\n", failures);
      deinitAll();
      return;
    }
  }

  /* Report running status */
  RDFDBServiceStatus.dwCurrentState = SERVICE_RUNNING; 
  RDFDBServiceStatus.dwCheckPoint   = 0; 
  RDFDBServiceStatus.dwWaitHint     = 0; 

  if ( !SetServiceStatus(RDFDBServiceStatusHandle, &RDFDBServiceStatus) ) { 
    PrintMsg(CRITICAL, "SetServiceStatus error %ld\n", GetLastError()); /* This one is critical because if it doesn't work RDFDB won't start fully. */
    return;
  } 

  /* Start the real work */
  openService(); 

  return; 
}

/* JvdG */
void WINAPI RDFDBServiceCtrlHandler(DWORD opcode)
{
  switch(opcode) { 
  case SERVICE_CONTROL_STOP: 
    RDFDBServiceStatus.dwWin32ExitCode = 0; 
    RDFDBServiceStatus.dwCurrentState  = SERVICE_STOP_PENDING; 
    RDFDBServiceStatus.dwCheckPoint    = 0; 
    RDFDBServiceStatus.dwWaitHint      = 0; 

    if ( !SetServiceStatus (RDFDBServiceStatusHandle, &RDFDBServiceStatus) ) { 
      PrintMsg(FAILURE, "SetServiceStatus error %ld\n", GetLastError()); 
    } 

    PrintMsg(INFO, "Exiting on Service Stop ...\n");

    {
      int failures;
      if ( 0 != (failures=deinitAll()) ) {
        PrintMsg(CRITICAL, "%i deinitializers failed.\n", failures);
      }
    }

    RDFDBServiceStatus.dwWin32ExitCode = 0; 
    RDFDBServiceStatus.dwCurrentState  = SERVICE_STOPPED; 
    RDFDBServiceStatus.dwCheckPoint    = 0; 
    RDFDBServiceStatus.dwWaitHint      = 0;
    break;
  case SERVICE_CONTROL_INTERROGATE:
    /* We're setting the status below anyhow, so we don't need to do anything in here. */
    break;
  }

  /* Send current status */
  if ( !SetServiceStatus(RDFDBServiceStatusHandle,  &RDFDBServiceStatus) ) { 
    PrintMsg(FAILURE, "SetServiceStatus error %ld\n", GetLastError()); 
  } 
}

/* JvdG */
/*! \todo Modify the commandline to list all options given to RDFDB, making sure the service option is set to run (if it was given through the commandline that is, if it wasn't the service option should be added). */
void CreateRDFDBService(int create, int argc, char* argv[])
{
  SC_HANDLE schService;
  SC_HANDLE schSCManager;

  schSCManager = OpenSCManager(
    0,  // pointer to machine name string
    0,  // pointer to database name string
    SC_MANAGER_CREATE_SERVICE);   // type of access

  if ( schSCManager == NULL ) {
    PrintMsg(CRITICAL, "Error opening service manager (%ld).\n", GetLastError());
    return;
  }

  schService = OpenService(
    schSCManager,
    TEXT("RDFDB"),
    SERVICE_ALL_ACCESS);

  if ( !create ) {
    /* We want to delete it */
    DWORD last_error = GetLastError();
    if ( schService == NULL ) {
      if ( last_error == ERROR_SERVICE_DOES_NOT_EXIST ) {
        PrintMsg(INFO, "The RDFDB service isn't installed (so I can't uninstall it).\n");
      }
      else {
        PrintMsg(CRITICAL, "Error opening the RDFDB service (%ld).\n", last_error);
      }
    }
    else {
      BOOL success = DeleteService(schService);
      if ( success ) {
        PrintMsg(INFO, "Successfully deleted the RDFDB service.\n");
      }
      else {
        PrintMsg(CRITICAL, "Error deleting the RDFDB service (%ld).\n", GetLastError());
      }
      CloseServiceHandle(schService); 
    }
  }
  else {
    char application_path[2048];
    char cmdline[2048];
    GetModuleFileName(0, application_path, (sizeof(application_path)/sizeof(application_path[0]))-1);
    snprintf(cmdline, 2047, "\"%s\" --service=run", application_path);
    cmdline[2047] = 0;

    {
      const char* const end = cmdline + sizeof(cmdline) - 1;
      char* cl = cmdline + strlen(cmdline);
      int i;
      for(i = 1; i < argc; ++i) {
        if ( !startsWith(argv[i], "--service") ) {
          const int al = strlen(argv[i]);
          if ( (end-cl) > al ) {
            *cl++ = ' ';
            strcpy(cl, argv[i]);
            cl += al;
          }
          else {
            PrintMsg(CRITICAL, "Commandline too long!\n");
            return;
          }
        }
      }
      *cl = 0;
    }

    if ( schService == NULL ) {
      /* Try creating it */
      schService = CreateService( 
        schSCManager,              // SCManager database 
        TEXT("RDFDB"),             // name of service 
        TEXT("RDFDB"),             // service name to display 
        SERVICE_ALL_ACCESS,        // desired access 
        SERVICE_WIN32_OWN_PROCESS, // service type 
        SERVICE_DEMAND_START,      // start type 
        SERVICE_ERROR_NORMAL,      // error control type 
        cmdline,          // service's binary 
        NULL,                      // no load ordering group 
        NULL,                      // no tag identifier 
        NULL,                      // no dependencies 
        NULL,                      // LocalSystem account 
        NULL);                     // no password

      if ( schService == NULL ) {
        PrintMsg(CRITICAL, "Error creating service (%ld).\n", GetLastError());
      }
      else {
        PrintMsg(INFO, "Successfully created the RDFDB service.\n");
        CloseServiceHandle(schService); 
      }
    }
    else {
      /* Opened it, we should now set the config */
      BOOL success = ChangeServiceConfig(
        schService,
        SERVICE_WIN32_OWN_PROCESS, // service type 
        SERVICE_DEMAND_START,      // start type 
        SERVICE_ERROR_NORMAL,      // error control type 
        cmdline,          // service's binary 
        NULL,                      // no load ordering group 
        NULL,                      // no tag identifier 
        NULL,                      // no dependencies 
        NULL,                      // LocalSystem account 
        NULL,                      // no password
        TEXT("RDFDB"));            // service name to display 

      if ( success ) {
        PrintMsg(INFO, "Successfully updated the RDFDB service configuration.\n");
      }
      else {
        PrintMsg(CRITICAL, "Error updating the RDFDB service configuration (%ld).\n", GetLastError());
      }

      CloseServiceHandle(schService); 
    }
  }
}
#endif