/* Copyright (c) 2004, Jasper van de Gronde
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "security.h"
#include "../utils/utils.h"
#include "../utils/hash.h"
#include <malloc.h>
#include <stdio.h>

struct _security_context_t {
  unsigned int ipv4address;
  unsigned int allowed_action_types;
};

typedef struct _user_info_t {
  const char* password;
  unsigned int allowed_action_types;
} user_info_t;
static user_info_t* new_user_info(const char* password, unsigned int allowed_action_types);
static void delete_user_info(user_info_t* user_info);

typedef struct _ipv4_info_t ipv4_info_t;
struct _ipv4_info_t {
  ipv4_info_t* next;
  ipv4_info_t* first_child;
  unsigned int begin;
  unsigned int end; /*!< Inclusive, so this value counts too. */
  unsigned int allowed_action_types;
};
static ipv4_info_t* new_ipv4_info(unsigned int begin, unsigned int end, unsigned int allowed_action_types);
static void delete_ipv4_info(ipv4_info_t* ipv4_info);

static HashTable users = 0;
static ipv4_info_t* ipv4ranges = 0;

/*! Returns 1 on success, 0 on failure. */
static int ipv4_set(unsigned int begin, unsigned end, unsigned int allowed_action_types);
/*! Returns the allowed_action_types for the specified address, 0 if the address isn't found. */
static unsigned int ipv4_get(unsigned int ipv4_address);

/*! Returns 1 on success, 0 on failure. */
static int parse_rights(const char* rights, unsigned int* parsed_rights);
/*! Returns a pointer to an internal buffer containing the serialized rights (use immediately!). */
static const char* serialize_rights(unsigned int rights);

/*! Returns 1 on success, 0 on failure. */
static int parse_ip_range(const char* rights, unsigned int* parsed_start_ip, unsigned int* parsed_stop_ip);
/*! Returns a pointer to an internal buffer containing the serialized rights (use immediately!). */
static const char* serialize_ip_range(unsigned int start_ip, unsigned int stop_ip);

int init_security() {
  int failures = 0;

  {
    const int old_failures = failures;
    const char* passwd = get_setting("USERS");
    FILE* passwd_file = 0;
    int users_added = 0;
    users = NewHashTable(64);
    if ( users && passwd && (passwd_file=fopen(passwd, "rt")) ) {
      char buffer[1024];
      PrintMsg(INFO, "Using users file: %s\n", passwd);
      while(!feof(passwd_file)) {
        if ( !fgets(buffer, sizeof(buffer), passwd_file) ) {
          if ( ferror(passwd_file) ) {
            PrintMsg(CRITICAL, "Error reading users file.\n");
            ++failures;
          }
          break;
        }
        inplace_trim_line(buffer);
        if ( *buffer ) {
          char* username = buffer;
          char* rights = username ? strchr(username, ':') : 0;
          char* passwd = rights ? strchr(rights+1, ':') : 0;
          unsigned int parsed_rights;
          if ( !passwd ) {
            PrintMsg(CRITICAL, "Syntax error in users file.\n");
            ++failures;
            break;
          }
          *rights = 0; ++rights;
          *passwd = 0; ++passwd;
          inplace_trim_line(username);
          inplace_trim_line(passwd);
          if ( strlen(username) == 0 ) {
            PrintMsg(CRITICAL, "Username must contain some non-whitespace characters.\n");
            ++failures;
            break;
          }
          
          if ( parse_rights(rights, &parsed_rights) ) {
            user_info_t* user_info = new_user_info(passwd, parsed_rights);
            if ( user_info ) {
              void* old_value = HashAdd(users, username, user_info);
              if ( old_value ) {
                delete_user_info(old_value);
                PrintMsg(CRITICAL, "User defined more than once: %s\n", username);
                ++failures;
                break;
              }
              ++users_added;
              PrintMsg(INFO, "User added: %s:%s\n", username, serialize_rights(parsed_rights));
            }
            else {
              PrintMsg(CRITICAL, "Error creating user_info.\n");
              ++failures;
              break;
            }
          }
          else {
            PrintMsg(CRITICAL, " (user=%s)\n", username);
            ++failures;
            break;
          }
        }
      }
      fclose(passwd_file);
      if ( users_added == 0 ) {
        PrintMsg(CRITICAL, "No users defined, this renders the server unuseable.\n");
        ++failures;
      }
    }
    else {
      if ( !users ) {
        PrintMsg(CRITICAL, "Error creating users hash table.\n");
      }

      if ( !passwd ) {
        PrintMsg(CRITICAL, "Users file not set.\n");
      }
      else if ( !passwd_file ) {
        PrintMsg(CRITICAL, "Unable to open users file: %s\n", passwd);
      }
      ++failures;
    }
    if ( failures != old_failures ) {
      DeleteHashTable(users);
      users = 0;
    }
  }
  
  {
    const int old_failures = failures;
    const char* ipfile_name = get_setting("IP_RANGES");
    FILE* ipfile = 0;
    int ranges_added = 0;
    if ( ipfile_name && (ipfile=fopen(ipfile_name, "rt")) ) {
      char buffer[1024];
      PrintMsg(INFO, "Using IP ranges file: %s\n", ipfile_name);
      while(!feof(ipfile)) {
        if ( !fgets(buffer, sizeof(buffer), ipfile) ) {
          if ( ferror(ipfile) ) {
            PrintMsg(CRITICAL, "Error reading IP ranges file.\n");
            ++failures;
          }
          break;
        }
        inplace_trim_line(buffer);
        if ( *buffer ) {
          char* ip_range = buffer;
          char* rights = ip_range ? strchr(ip_range, ':') : 0;
          unsigned int parsed_start_ip, parsed_stop_ip;
          unsigned int parsed_rights;
          if ( !rights ) {
            PrintMsg(CRITICAL, "Syntax error in users file.\n");
            ++failures;
            break;
          }
          *rights = 0; ++rights;
          inplace_trim_line(ip_range);
          
          if ( parse_ip_range(ip_range, &parsed_start_ip, &parsed_stop_ip)
            && parse_rights(rights, &parsed_rights) )
          {
            if ( ipv4_set(parsed_start_ip, parsed_stop_ip, parsed_rights) ) {
              ++ranges_added;
              PrintMsg(INFO, "IP range added: %s:%s\n", serialize_ip_range(parsed_start_ip, parsed_stop_ip), serialize_rights(parsed_rights));
            }
            else {
              PrintMsg(CRITICAL, "Error setting IP range: %s:%s\n", serialize_ip_range(parsed_start_ip, parsed_stop_ip), serialize_rights(parsed_rights));
              ++failures;
              break;
            }
          }
          else {
            PrintMsg(CRITICAL, "(IP range=%s)\n", ip_range);
            ++failures;
            break;
          }
        }
      }
      fclose(ipfile);
      if ( ranges_added == 0 ) {
        PrintMsg(CRITICAL, "No IP ranges defined, this renders the server unuseable.\n");
        ++failures;
      }
    }
    else {
      if ( !ipfile_name ) {
        PrintMsg(CRITICAL, "IP ranges file not set.\n");
      }
      else if ( !ipfile ) {
        PrintMsg(CRITICAL, "Unable to open IP ranges file: %s\n", ipfile_name);
      }
      ++failures;
    }
    if ( failures != old_failures ) {
      delete_ipv4_info(ipv4ranges);
      ipv4ranges = 0;
    }
  }
  /* For testing purposes */
  /*ipv4_set(0, -1, -1);*/

  return failures;
}

int deinit_security() {
  int failures = 0;

  if ( users ) {
    int i;
    for(i = 0; i < users->size; ++i) {
      HashEntry entry = users->buckets[i];
      while(entry) {
        delete_user_info(entry->value);
        entry = entry->next;
      }
    }
    DeleteHashTable(users);
    users = 0;
  }

  delete_ipv4_info(ipv4ranges);
  ipv4ranges = 0;

  return failures;
}

security_context_t* new_security_context(unsigned int ipv4address) {
  security_context_t* security_context = malloc(sizeof(security_context_t));
  if ( security_context ) {
    memset(security_context, 0, sizeof(security_context_t));
    security_context->ipv4address = ipv4address;
    security_context->allowed_action_types = action_connect & ipv4_get(security_context->ipv4address);
  }
  return security_context;
}

void delete_security_context(security_context_t* security_context) {
  if ( security_context ) {
    free(security_context);
  }
}

int authenticate(security_context_t* security_context, const identity_t* identity) {
  int global_user_permissions;
  int global_ip_permissions;
  if ( !security_context || !identity || !identity->username || !identity->password ) return 0;

  {
    user_info_t* user_info = HashLookup(users, identity->username);
    if ( !user_info || 0 != strcmp(identity->password, user_info->password) ) {
      return 0;
    }
    global_user_permissions = user_info->allowed_action_types;
  }

  global_ip_permissions = ipv4_get(security_context->ipv4address);

  security_context->allowed_action_types = global_user_permissions & global_ip_permissions;

  return 1;
}

int action_allowed(security_context_t* security_context, unsigned int action_types, const action_info_t* action_info) {
  if ( !security_context ) return 0;

  if ( action_types != (action_types & security_context->allowed_action_types) ) {
    return 0;
  }

  return 1;
}

static user_info_t* new_user_info(const char* password, unsigned int allowed_action_types) {
  user_info_t* user_info;
  if ( !password ) return 0;
  user_info = malloc(sizeof(user_info_t));
  if ( user_info ) {
    user_info->password = strdup(password);
    user_info->allowed_action_types = allowed_action_types;
  }
  return user_info;
}

static void delete_user_info(user_info_t* user_info) {
  if ( user_info ) {
    free((void*)(user_info->password));
    free(user_info);
  }
}

static ipv4_info_t* new_ipv4_info(unsigned int begin, unsigned int end, unsigned int allowed_action_types) {
  ipv4_info_t* ipv4_info;
  if ( begin > end ) return 0;
  ipv4_info = malloc(sizeof(ipv4_info_t));
  if ( ipv4_info ) {
    ipv4_info->next = 0;
    ipv4_info->first_child = 0;
    ipv4_info->begin = begin;
    ipv4_info->end = end;
    ipv4_info->allowed_action_types = allowed_action_types;
  }
  return ipv4_info;
}

static void delete_ipv4_info(ipv4_info_t* ipv4_info) {
  ipv4_info_t* next;
  while(ipv4_info) {
    delete_ipv4_info(ipv4_info->first_child);
    next = ipv4_info->next;
    free(ipv4_info);
    ipv4_info = next;
  }
}

static ipv4_info_t* ipv4_lookup(unsigned int begin, unsigned end) {
  ipv4_info_t* last_match = 0;
  ipv4_info_t* ipv4_info = ipv4ranges;
  while(ipv4_info) {
    if ( begin > ipv4_info->end ) {
      ipv4_info = ipv4_info->next;
    }
    else if ( end < ipv4_info->begin ) {
      return last_match;
    }
    else if ( begin == ipv4_info->begin && end == ipv4_info->end ) {
      return ipv4_info;
    }
    else if ( begin >= ipv4_info->begin && end <= ipv4_info->end ) {
      last_match = ipv4_info;
      ipv4_info = ipv4_info->first_child;
    }
    else {
      return last_match;
    }
  }
  return last_match;
}

static int ipv4_set(unsigned int begin, unsigned end, unsigned int allowed_action_types) {
  ipv4_info_t* ipv4_info;
  ipv4_info_t* parent;
  ipv4_info_t* prev = 0;

  if ( !ipv4ranges ) {
    ipv4ranges = new_ipv4_info(begin, end, allowed_action_types);
    if ( ipv4ranges ) {
      return 1;
    }
  }

  ipv4_info = ipv4_lookup(begin, end);
  if ( ipv4_info && begin == ipv4_info->begin && end == ipv4_info->end ) {
    ipv4_info->allowed_action_types = allowed_action_types;
    return 1;
  }

  if ( ipv4_info ) {
    if ( !ipv4_info->first_child ) {
      ipv4_info->first_child = new_ipv4_info(begin, end, allowed_action_types);
      if ( ipv4_info->first_child ) {
        return 1;
      }
    }
    parent = ipv4_info;
    ipv4_info = ipv4_info->first_child;
  }
  else {
    parent = 0;
    ipv4_info = ipv4ranges;
  }

  while(ipv4_info) {
    if ( end < ipv4_info->begin ) {
      ipv4_info_t* ii = new_ipv4_info(begin, end, allowed_action_types);
      if ( ii ) {
        ii->next = ipv4_info;
        if ( prev ) {
          prev->next = ii;
        }
        else if ( parent ) {
          parent->first_child = ii;
        }
        else {
          ipv4ranges = ii;
        }
        return 1;
      }
      else {
        return 0;
      }
    }
    else if ( begin <= ipv4_info->begin && end >= ipv4_info->end ) {
      ipv4_info_t* ii = new_ipv4_info(begin, end, allowed_action_types);
      if ( ii ) {
        ipv4_info_t* p = 0;
        ii->first_child = ipv4_info;
        while(ipv4_info && end >= ipv4_info->end ) {
          p = ipv4_info;
          ipv4_info = ipv4_info->next;
        }
        if ( ipv4_info && end >= ipv4_info->begin ) {
          delete_ipv4_info(ii);
          return 0;
        }
        if ( p ) {
          p->next = 0;
        }
        ii->next = ipv4_info;
        if ( prev ) {
          prev->next = ii;
        }
        else if ( parent ) {
          parent->first_child = ii;
        }
        else {
          ipv4ranges = ii;
        }
        return 1;
      }
      else {
        return 0;
      }
    }
    else if ( begin <= ipv4_info->end ) {
      return 0;
    }
    else if ( ipv4_info->next == 0 ) {
      ipv4_info->next = new_ipv4_info(begin, end, allowed_action_types);
      if ( ipv4_info->next ) {
        return 1;
      }
    }
    prev = ipv4_info;
    ipv4_info = ipv4_info->next;
  }

  return 0;
}

static unsigned int ipv4_get(unsigned int ipv4_address) {
  ipv4_info_t* ipv4_info = ipv4_lookup(ipv4_address, ipv4_address);
  if ( ipv4_info ) {
    return ipv4_info->allowed_action_types;
  }
  else {
    return 0;
  }
}

/* Any error messages printed by this function do not end with a newline, as its caller will append some information. */
static int parse_rights(const char* rights, unsigned int* parsed_rights) {
  unsigned int all = 0;
  unsigned int ps = 0;
  *parsed_rights = 0;
  while(*rights) {
    const char* begin = rights;
    const char* end;
    while(isspace(*begin)) ++begin;
    if ( !*begin ) break;
    end = begin;
    while(*end && !isspace(*end)) ++end;

    if ( all == 1 ) {
      if ( strcmprange("except", begin, end) == 0 ) {
        all = 2;
      }
      else {
        PrintMsg(CRITICAL, "The 'all' permission should not be followed by anything other than an optional 'except' with a list of permissions behind it.");
        return 0;
      }
    }
    else if ( strcmprange("connect", begin, end) == 0 ) {
      ps |= action_connect;
    }
    else if ( strcmprange("admin", begin, end) == 0 ) {
      ps |= action_admin;
    }
    else if ( strcmprange("read_db", begin, end) == 0 ) {
      ps |= action_read_db;
    }
    else if ( strcmprange("write_db", begin, end) == 0 ) {
      ps |= action_write_db;
    }
    else if ( strcmprange("read_url", begin, end) == 0 ) {
      ps |= action_read_url;
    }
    else if ( strcmprange("write_url", begin, end) == 0 ) {
      ps |= action_write_url;
    }
    else if ( strcmprange("all", begin, end) == 0 ) {
      if ( ps == 0 && all == 0 ) {
        all = 1;
      }
      else {
        PrintMsg(CRITICAL, "The 'all' permission is only valid as the first permission (optionally followed by 'except' and a list of permissions that aren't allowed).");
        return 0;
      }
    }
    else {
      char buffer[1024];
      strncpy(buffer, begin, end-begin);
      PrintMsg(CRITICAL, "Unrecognized permission: %s.", buffer);
      return 0;
    }
    rights = end;
  }
  if ( all ) {
    ps = ~ps;
  }
  *parsed_rights = ps;
  return 1;
}

static const char* serialize_rights(unsigned int rights) {
  static char buffer[1024];
  buffer[0] = 0;
  if ( (rights & action_connect) ) {
    strcat(buffer, "connect ");
  }
  if ( (rights & action_admin) ) {
    strcat(buffer, "admin ");
  }
  if ( (rights & action_read_db) ) {
    strcat(buffer, "read_db ");
  }
  if ( (rights & action_write_db) ) {
    strcat(buffer, "write_db ");
  }
  if ( (rights & action_read_url) ) {
    strcat(buffer, "read_url ");
  }
  if ( (rights & action_write_url) ) {
    strcat(buffer, "write_url ");
  }
  if ( *buffer && buffer[strlen(buffer)-1] == ' ' ) {
    buffer[strlen(buffer)-1] = 0;
  }
  return buffer;
}

static const char* parse_ip(const char* ip_range, unsigned int* parsed_ip, unsigned int* wildcard) {
  unsigned int num;

  *parsed_ip = 0;
  *wildcard = 0;

  num = strtoul(ip_range, &((char*)ip_range), 10);
  if ( *ip_range == '*' ) { *wildcard = 0xffffffff; return ip_range+1; }
  if ( num > 255 ) return 0;
  *parsed_ip |= num << 24;
  if ( *ip_range != '.' ) return 0; else ++ip_range;

  num = strtoul(ip_range, &((char*)ip_range), 10);
  if ( *ip_range == '*' ) { *wildcard = 0x00ffffff; return ip_range+1; }
  if ( num > 255 ) return 0;
  *parsed_ip |= num << 16;
  if ( *ip_range != '.' ) return 0; else ++ip_range;

  num = strtoul(ip_range, &((char*)ip_range), 10);
  if ( *ip_range == '*' ) { *wildcard = 0x0000ffff; return ip_range+1; }
  if ( num > 255 ) return 0;
  *parsed_ip |= num << 8;
  if ( *ip_range != '.' ) return 0; else ++ip_range;

  num = strtoul(ip_range, &((char*)ip_range), 10);
  if ( *ip_range == '*' ) { *wildcard = 0x000000ff; return ip_range+1; }
  if ( num > 255 ) return 0;
  *parsed_ip |= num;

  return ip_range;
}

static void ip_to_string(char* buffer, unsigned int ip) {
  sprintf(buffer, "%i.%i.%i.%i", (ip>>24)&255, (ip>>16)&255, (ip>>8)&255, (ip>>0)&255);
}

/* Any error messages printed by this function do not end with a newline, as its caller will append some information. */
static int parse_ip_range(const char* ip_range_p, unsigned int* parsed_start_ip, unsigned int* parsed_stop_ip) {
  const char* ip_range = ip_range_p;
  unsigned int wildcard;
  while(isspace(*ip_range)) ++ip_range;
  ip_range = parse_ip(ip_range, parsed_start_ip, &wildcard);
  *parsed_stop_ip = *parsed_start_ip;
  if ( ip_range ) {
    while(isspace(*ip_range)) ++ip_range;
    if ( *ip_range == '-' ) {
      if ( wildcard ) {
        PrintMsg(CRITICAL, "An IP address with a wildcard in it can't be part of a range.");
        ip_range = 0;
      }
      else {
        ++ip_range;
        while(isspace(*ip_range)) ++ip_range;
        ip_range = parse_ip(ip_range, parsed_stop_ip, &wildcard);
        if ( wildcard ) {
          PrintMsg(CRITICAL, "An IP address with a wildcard in it can't be part of a range.");
          ip_range = 0;
        }
      }
    }
    else if ( wildcard ) {
      *parsed_stop_ip |= wildcard;
    }
    if ( ip_range ) {
      while(isspace(*ip_range)) ++ip_range;
      if ( *ip_range ) {
        PrintMsg(CRITICAL, "Unrecognized tail in IP range: %s.", ip_range);
        ip_range = 0;
      }
    }
  }
  else {
    PrintMsg(CRITICAL, "Syntax error in IP range: %s.", ip_range_p);
  }
  return ip_range ? 1 : 0;
}

static const char* serialize_ip_range(unsigned int start_ip, unsigned int stop_ip) {
  static char buffer[1024];
  ip_to_string(buffer, start_ip);
  if ( stop_ip != start_ip ) {
    strcat(buffer, "-");
    ip_to_string(buffer+strlen(buffer), stop_ip);
  }
  return buffer;
}
