/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors: Ivan Kurmanov, Jasper van de Gronde (win32 port)
 */

#ifdef WIN32
#include <winsock2.h>
#include <windows.h>
#endif
#include "utils.h"
#include "hash.h"
#include <stdio.h>

/* JvdG Replacement of gVerbose */
#ifdef _DEBUG
unsigned int verbosity_level_mask = -1;
#else
unsigned int verbosity_level_mask = NOTICE|CRITICAL|INFO;
#endif

/* JvdG */
/** Stores already retrieved settings. */
HashTable settings = 0;

static FILE* logfile = 0;

static int store_setting(const char* str, char** key, size_t* cur_key_size, int overwrite);

int deinitUtils();

/* JvdG */
int initUtils(int argc, char* argv[]) {
  int failures = 0;

  deinitUtils();

  { /* Set up the settings hash table */
    size_t cur_key_size = 0;
    char* key = 0;

    settings = NewHashTable(256);

    {
      int i;
      for(i = 1; i < argc; ++i) { /* Start at 1 to skip the executable's name. */
        if ( startsWith(argv[i], "--") ) {
          if ( 0 != store_setting(argv[i]+2, &key, &cur_key_size, 1) ) {
            break;
          }
        }
      }
    }

    {
      const char* const config_filename = get_setting("CONFIG_FILE");
      if ( config_filename ) {
        FILE* config_file = fopen(config_filename, "rt");
        if ( config_file ) {
          char org_buffer[2048];
          const size_t buffer_size = sizeof(org_buffer)/sizeof(org_buffer[0]);
          char* buffer = 0;
          size_t bytes_in_buffer = 0;
          for(;;) {
            char* end_of_line = (char*)strnchr(buffer, 10, bytes_in_buffer);
            if ( !end_of_line ) {
              if ( bytes_in_buffer < buffer_size ) {
                size_t bytes_read;
                memmove(org_buffer, buffer, bytes_in_buffer);
                bytes_read = fread(org_buffer+bytes_in_buffer, sizeof(org_buffer[0]), buffer_size-bytes_in_buffer, config_file);
                if ( bytes_read < buffer_size-bytes_in_buffer && bytes_in_buffer+bytes_read > 0 ) {
                  org_buffer[bytes_in_buffer+bytes_read] = 10;
                  ++bytes_read;
                }
                bytes_in_buffer += bytes_read;
                buffer = org_buffer;
                if ( bytes_in_buffer > 0 ) {
                  continue;
                }
                else {
                  break;
                }
              }
              else {
                break;
              }
            }
            *end_of_line = 0;

            if ( 0 != store_setting(buffer, &key, &cur_key_size, 0) ) {
              break;
            }

            bytes_in_buffer -= (end_of_line+1)-buffer;
            buffer = end_of_line+1;
          } /* while(!feof(config_file)) */
          fclose(config_file);
        }
        else {
          PrintMsg(CRITICAL, "I can't open the file specified as config file (%s).\n", config_filename);
          ++failures;
        }
      }
    }

    if ( key ) free(key);
  } /* Finished setting up the settings hash table */

  {
    const char* const logfile_filename = get_setting("LOG_FILE");
    if ( logfile_filename ) {
      logfile = fopen(logfile_filename, "a+");
      if ( !logfile ) {
        PrintMsg(CRITICAL, "I can't open the file specified as log file (%s).\n", logfile_filename);
        ++failures;
      }
    }
  }

  {
    char* end_of_conv = 0;
    const char* const level_mask_str = get_setting("VERBOSITY_MASK");
    if ( level_mask_str ) {
      unsigned int level_mask = strtoul(level_mask_str, &end_of_conv, 16);
      if ( *end_of_conv != 0 || (level_mask & 0xff) != 0) {
        PrintMsg(CRITICAL, "The verbosity mask was set to an illegal value (%s).\n", level_mask_str);
        ++failures;
      }
      else {
        verbosity_level_mask = level_mask;
      }
    }
  }
  return failures;
}

/* JvdG */
int deinitUtils() {
#ifdef _DEBUG
  verbosity_level_mask = -1;
#else
  verbosity_level_mask = NOTICE|CRITICAL|INFO;
#endif
  if ( logfile ) {
    fclose(logfile);
    logfile = 0;
  }
  if ( settings ) {
    int i;
    for(i = 0; i < settings->size; ++i) {
      HashEntry he = settings->buckets[i];
      while(he) {
        freeMem(he->value);
        he = he->next;
      }
    }
    DeleteHashTable(settings);
    settings = 0;
  }
  return 0;
}

/* JvdG */
static int store_setting(const char* str, char** key, size_t* cur_key_size, int overwrite) {
  const char* const equals_pos = strchr(str, '=');
  if ( equals_pos ) {
    const size_t key_size = (equals_pos-str)+1;
    if ( key_size <= 1 ) return 0;
    if ( key_size > *cur_key_size ) {
      *cur_key_size = key_size;
      *key = realloc(*key, *cur_key_size);
      if ( !*key ) return 1;
    }
    touppern(*key, str, key_size);
    if ( overwrite ) {
      char* const value = strdup(equals_pos+1);
      void* old_value = HashAdd(settings, *key, value);
      if ( old_value ) free(old_value);
    }
    else {
      void* exists = HashLookup(settings, *key);
      if ( !exists ) {
        char* const value = strdup(equals_pos+1);
        HashAdd(settings, *key, value);
      }
    }
  }
  return 0;
}

/* JvdG */
const char* get_setting(const char* key) {
  char* value = HashLookup(settings, key);
  if ( value == 0 ) {
    const size_t env_var_size = strlen(key) + 6 + 1;
    char* env_var = malloc(env_var_size);
    if ( env_var ) {
      strcpy(env_var, "RDFDB_");
      touppern(env_var+6, key, env_var_size-1);
      value = getenv(env_var);
      free(env_var);
      if ( value ) {
        value = strdup(value);
        HashAdd(settings, key, value);
      }
    }
  }
  return value;
}

char* 
getMem (size_t l) {
  char* ans = malloc(l);
  if (ans) bzero(ans, l);
  return ans;
}

void
freeMem (const void* mem) {
  free((void*)mem); 
}

int reserve(void** mem, size_t* cur_size, size_t requested_size) {
  if ( requested_size > *cur_size ) {
    void* memory = realloc(*mem, requested_size);
    if ( !memory ) return 0;
    *mem = memory;
    *cur_size = requested_size;
  }
  return 1;
}

char* 
trimLine (char* line) {
  size_t len = strlen(line);
  size_t index = len-1;
  while (index > 0) {
    if (isspace(line[index])) {
      line[index] = 0;
      index--;
    } else break;
  }
  index = 0;
  while (index < len) {
    if (isspace(line[index])) {
      index++;
    } else break;
  }
  return &line[index];
}

/* JvdG */
int inplace_trim_line(char* line) {
  int removed = 0;
  char* begin = line;
  char* end;
  char* cur;
  while(isspace(*begin)) ++begin;
  end = cur = begin;
  while(*cur) {
    while(*cur && !isspace(*cur)) ++cur;
    end = cur;
    while(isspace(*cur)) ++cur;
  }
  memmove(line, begin, end-begin);
  *(line+(end-begin)) = 0;
  removed = (cur-line)-(end-begin);
  return removed;
}

/*
int
nullString (const char* line, int len) {
  int n = 0;
  while (n < len) {
    if (!isspace(line[n])) return 0;
    n++;
  } 
  return 1;
}
*/

int
startsWith (const char* str, const char* prefix) {
  size_t l1 = strlen(str);
  size_t l2 = strlen(prefix);
  if (l1 >= l2) {
    return (strncasecmp(str, prefix, l2) == 0);
  } else return 0;
}


/*****   starts with forgiving white space    ******/
int
startsWith_forgiving_ws (const char* actual_str, const char* prefix) {
  const char * str = actual_str;
  while( isspace( str[0] ) ) {
    str++;
  }

  {
    size_t l1 = strlen(str);
    size_t l2 = strlen(prefix);
    if (l1 >= l2) {
      return (strncasecmp(str, prefix, l2) == 0);
    } else return 0;
  }
}

int
endsWith (const char* str, const char* suffix) {
  size_t l1 = strlen(str);
  size_t l2 = strlen(suffix);

  if (l1 >= l2) {
    return (strncasecmp(&str[l1-l2], suffix, l2) == 0);
  } else return 0;
}

/* Jasper van de Gronde
 * I made this to calculate how many digits a number will take,
 * so this can be used in calculating how much memory must be
 * reserved when converting the number to a string.
 */
int numDigits(int n)
{
    int digits;
    if ( n < 0 ) {
        digits = 2; /* take the '-' into account */
        n = -n;
    } else {
        digits = 1;
    }
    while(n > 9) {
        n /= 10;
        ++digits;
    }
    return digits;
}

/* Jasper van de Gronde
 * Duplicates the string starting at begin and ending at end (not including end).
 */
char* strduprange(const char* begin, const char* end) {
  const size_t size = end-begin;
  char* s = getMem(size+1);
  if ( s ) {
    memcpy(s, begin, size);
    /* No need to explicitly set s[size] to 0, as getMem initializes the memory to 0. */
  }
  return s;
}

/* Jasper van de Gronde
 * Compares s with the string starting at begin and ending at end (not including end)
 */
int strcmprange(const char* s, const char* begin, const char* end) {
  while(*s && begin != end && *s == *begin) {
    ++s;
    ++begin;
  }
  if ( !*s && begin == end ) return 0;
  if ( !*s && begin != end ) return -1;
  if ( *s && begin == end ) return 1;
  if ( *s < *begin ) return -1;
  if ( *s > *begin ) return 1;
  return 0;
}

/* JvdG */
/** Concatenates s1, c and s2. */
char* concat(const char* s1, const char* s2) {
  size_t s1_len = strlen(s1);
  char* const str = getMem(s1_len + strlen(s2) + 1);
  if ( str ) {
    memcpy(str, s1, s1_len);
    strcpy(str+s1_len, s2);
  }
  return str;
}

/* JvdG */
void touppern(char* dest, const char* src, size_t dest_size) {
  size_t i;
  for(i = 0; i < dest_size-1 && *src; ++i) {
    *dest++ = toupper(*src++);
  }
  *dest = 0;
}

/* JvdG */
const char* strnchr(const char* s, char ch, size_t n) {
  const char* end = s+n;
  while(s!=end && *s!=ch) ++s;
  if ( s == end ) {
    return 0;
  }
  else {
    return s;
  }
}

/* JvdG */
char*
outputEscapedString(const char* str, char* line) {
  *line++ = '"';
  while(*str) {
    switch(*str) {
    case '\xA':
      *line++ = '\\';
      *line++ = 'n';
      break;
    case '\xD':
      *line++ = '\\';
      *line++ = 'r';
      break;
    case '\x9':
      *line++ = '\\';
      *line++ = 't';
      break;
    case '"':
    case '\\':
      *line++ = '\\';
    default:
      *line++ = *str;
    }
    ++str;
  }
  *line++ = '"';
  *line = 0;
  return line;
}

/* JvdG */
char*
outputCSVEscapedString(const char* str, char* line) {
  *line++ = '"';
  while(*str) {
    switch(*str) {
    case '"':
      *line++ = '"';
    default:
      *line++ = *str;
    }
    ++str;
  }
  *line++ = '"';
  *line = 0;
  return line;
}

/* JvdG */
/** Prints a message if the level is enabled according to IsLevelEnabled. */
int PrintMsg(VerbosityLevel level, const char *format, ...) {
  va_list arglist;
  va_start(arglist, format);
  if ( IsLevelEnabled(level) ) {
    if ( logfile ) {
      vfprintf(logfile, format, arglist);
      fflush(logfile);
    }
    return vprintf(format, arglist);
  }
  else {
    return 0;
  }
}

/* JvdG */
/** Prints a message if the level is enabled according to IsLevelEnabled. */
int IsLevelEnabled(VerbosityLevel level) {
  return (level & verbosity_level_mask) != 0;
}
